@@ -8,6 +8,7 @@ from ast import literal_eval
 from typing import Union, List, Tuple
 
 from faker import Faker
+from faker.providers.date_time import Provider as DateProvider
 
 from .data_gen_exceptions import DataGenError
 
@@ -23,7 +24,7 @@ fake = Faker()
 # Python 3.6 is out of the support matrix.
 
 
-def parse_weight_str(context: PluginContext, weight_value) -> int:
+def parse_weight_str(context: PluginContext, weight_value) -> float:
     """For constructs like:
 
     - choice:
@@ -35,10 +36,10 @@ def parse_weight_str(context: PluginContext, weight_value) -> int:
     weight_str = context.evaluate(weight_value)
     if isinstance(weight_str, str):
         weight_str = weight_str.rstrip("%")
-    return int(weight_str)
+    return float(weight_str)
 
 
-def weighted_choice(choices: List[Tuple[int, object]]):
+def weighted_choice(choices: List[Tuple[float, object]]):
     """Selects from choices based on their weights"""
     weights = [weight for weight, value in choices]
     options = [value for weight, value in choices]
@@ -47,10 +48,11 @@ def weighted_choice(choices: List[Tuple[int, object]]):
 
 @lru_cache(maxsize=512)
 def parse_date(d: Union[str, datetime, date]) -> date:
-    if isinstance(d, date):
-        return d
-    elif isinstance(d, datetime):
+    if isinstance(d, datetime):
         return d.date()
+    elif isinstance(d, date):
+        return d
+
     return dateutil.parser.parse(d).date()
 
 
@@ -96,15 +98,17 @@ class StandardFuncs(SnowfakeryPlugin):
 
         def date_between(self, *, start_date, end_date):
             """A YAML-embeddable function to pick a date between two ranges"""
-            try:
-                start_date = parse_date(start_date)
-            except Exception:  # let's hope its something faker can parse
-                pass
 
-            try:
-                end_date = parse_date(end_date)
-            except Exception:  # let's hope its something faker can parse
-                pass
+            def try_parse_date(d):
+                if not isinstance(d, str) or not DateProvider.regex.fullmatch(d):
+                    try:
+                        d = parse_date(d)
+                    except Exception:  # let's hope its something faker can parse
+                        pass
+                return d
+
+            start_date = try_parse_date(start_date)
+            end_date = try_parse_date(end_date)
 
             try:
                 return fake.date_between(start_date, end_date)
@@ -118,17 +122,19 @@ class StandardFuncs(SnowfakeryPlugin):
             func = getattr(faker, fake)
             return func()
 
-        def random_number(self, min: int, max: int) -> int:
+        def random_number(self, min: int, max: int, step: int = 1) -> int:
             """Pick a random number between min and max like Python's randint."""
-            return random.randint(min, max)
+            return random.randrange(min, max + 1, step)
 
         def reference(self, x: Union[ObjectRow, str]):
             """YAML-embeddable function to Reference another object."""
             if hasattr(x, "id"):  # reference to an object with an id
                 target = x
             elif isinstance(x, str):  # name of an object
-                obj = self.context.field_vars()[x]
-                if not getattr(obj, "id"):
+                obj = self.context.field_vars().get(x)
+                if not obj:
+                    raise DataGenError(f"Cannot find an object named {x}", None, None)
+                if not getattr(obj, "id", None):
                     raise DataGenError(
                         f"Reference to incorrect object type {obj}", None, None
                     )
@@ -141,7 +147,7 @@ class StandardFuncs(SnowfakeryPlugin):
             return target
 
         @lazy
-        def random_choice(self, *choices):
+        def random_choice(self, *choices, **kwchoices):
             """Template helper for random choices.
 
             Supports structures like this:
@@ -168,16 +174,33 @@ class StandardFuncs(SnowfakeryPlugin):
 
             Pick-items are lazily evaluated.
             """
-            if not choices:
-                raise ValueError("No choices supplied!")
 
-            if getattr(choices[0], "function_name", None) == "choice":
-                choices = [self.context.evaluate(choice) for choice in choices]
-                rc = weighted_choice(choices)
+            # very occasionally single-item choices are useful
+            use_choices = len(choices) >= 1
+
+            # very occasionally single-item choices are useful
+            use_kwchoices = len(kwchoices) >= 1
+
+            if not (use_choices or use_kwchoices):
+                raise ValueError("No choices supplied!")
+            elif use_choices and use_kwchoices:
+                raise ValueError("Both choices and probabilities supplied!")
+            elif use_choices:
+                if getattr(choices[0], "function_name", None) == "choice":
+                    choices = [self.context.evaluate_raw(choice) for choice in choices]
+                    rc = weighted_choice(choices)
+                else:
+                    rc = random.choice(choices)
+                if hasattr(rc, "render"):
+                    rc = self.context.evaluate_raw(rc)
             else:
-                rc = random.choice(choices)
-            if hasattr(rc, "render"):
-                rc = self.context.evaluate(rc)
+                assert use_kwchoices and not use_choices
+                choices = [
+                    (parse_weight_str(self.context, value), key)
+                    for key, value in kwchoices.items()
+                ]
+                rc = weighted_choice(choices)
+
             return rc
 
         @lazy
@@ -213,7 +236,7 @@ class StandardFuncs(SnowfakeryPlugin):
             if not choices:
                 raise ValueError("No choices supplied!")
 
-            choices = [self.context.evaluate(choice) for choice in choices]
+            choices = [self.context.evaluate_raw(choice) for choice in choices]
             for when, choice in choices[:-1]:
                 if when is None:
                     raise SyntaxError(
@@ -226,7 +249,7 @@ class StandardFuncs(SnowfakeryPlugin):
             )
             rc = next(true_choices, choices[-1][-1])  # default to last choice
             if hasattr(rc, "render"):
-                rc = self.context.evaluate(rc)
+                rc = self.context.evaluate_raw(rc)
             return rc
 
     setattr(Functions, "if", Functions.if_)
